#!/usr/bin/env python

##############################################################
#                     (C) SANDULEAC DAN                      #
#                  EVALUATOR   version 0.64                  #
#                  <<  Console interface  >>                 #
#                 RELEASED UNDER GPL LICENCE                 #
##############################################################

import evalcore
import sys, os, re
from os.path import join, isfile, isdir


ESC = chr(27)
color = {
    'green' : ESC+'[1;32m',
    'green_i' : ESC+'[0;32m',
    'red'   : ESC+'[1;31m',
    'else'  : ESC+'[1;35m',
    'maialb': ESC+'[1;37m',
    'normal': ESC+'[m'
}

def col(c): return color[c]
write = sys.stdout.write

log_print = evalcore.log_print
die = evalcore.die

# Cu tine ce fac??

def setup_output(quiet = False, output_path = ''):
    output = sys.stdout
    if quiet:
        output = None
    # stdout
    if output_path == '-':
        output = sys.stdout
    elif output_path:
        try:
            output = open(output_path, 'wt')
        except IOError, e:
            print >>sys.stderr, "Can't open desired output file:"
            raise e
    return output
    # TODO : TOT AICI
    # set global verbosity according to type of output and parsed verbosity


def parse_parameters(argv):
    ''' Parses shell parameters
    Returns tuple (params, options)
    `Params` are global or one-time parameters
    `options` are problem-specific parameters
    '''
    from optparse import OptionParser
    "Parses parameters; uses `defaults` and saves to dict `options`"
    parser = OptionParser(usage="%prog executabil [options]")
    rules = [
        (['-p', '--prob'], {'metavar':'...', 'dest':'path', 'help':'Numele problemei (si directorului de teste)'}),
        (['-n', '--name'], {'dest':'nume', 'help':'numele problemei in teste, daca diferit de -p'}),
        (['-d'],           {'dest':'debug', 'action':"store_true", 'default':False, 'help':'debug pentru testele picate'}),
        (['-t', '--time'], {'type':'int', 'help':'time limit in milisecunde'}),
        (['-m', '--memory'], {'metavar':'MEM', 'type':"int", 'help':'limita memorie, kb'}),
        (['-q', '--quiet'], {'dest':'quiet', 'action':'store_true', 'default':False, 'help':'Nu afiseaza decat scorul'}),
        (['-O'], {'dest':'output', 'help':'Fisier pentru salvarea borderoului'}),
        (['--chroot'],     {'action':'store_true', 'default':False,
                            'help':'''Foloseste chroot (jrun are setuid/esti root)
                            Comanda pt jrun: %s
                            ''' % evalcore.eval_options['chroot_command']}),
        (['-r', '--repo'], {'action': 'store_true', 'default':False, 'help':'intotdeauna foloseste repository'}),
        (['-i', '--inf'],  {'help':'''Format fisier intrare (default %default)
                                   %s se expandeaza ca numele problemei, ? ca nr testului'''}),
        (['-o', '--outf'], {'help':'Format fisier iesire'}),
        (['--verif'],      {'help':'''Program verificator
                                   Parametri: VERIF fisier.in fisier.out fisier.ok
                                   Returneaza: "mesaj\\n(procent punctaj din max)"
                                   Exemplu: Gresit\\n0 ; sau: OK\\n100'''})
    ]
    for i, j in rules: parser.add_option(*i, **j)
    parser.set_defaults(**evalcore.defaults)
    (opts, args) = parser.parse_args(argv)
    if len(args) < 2:
        die("Trebuie sa precizezi numele executabilului pe care-l evaluez")
    if len(args) > 2:
        die("Parametru in plus '%s'" % args[2])


    params_list = ('debug', 'repo', 'quiet', 'output', 'chroot')
    params = dict((i, opts.__dict__[i]) for i in params_list)
    options_list = ('inf', 'outf', 'nume', 'path', 'memory', 'time', 'verif')
    options = dict((i, opts.__dict__[i]) for i in options_list)

    params['executabil'] = re.sub('^./', '', args[1])

    return (params, options)


def print_results(problema, executabil, params = {}, **kwargs):
    ''' Main show() function for the console interface
        **kwargs -> only among: ['debug' (Bool), ]
    '''
    tester = problema.evaluate(executabil, **kwargs)
    display_table = hasattr(evalcore, 'output') and evalcore.output

    lens = [12, 10, 10, 10, 10, 10]
    score = real_score = 0
    nr_teste = 0

    visual_notice =__name__ == '__main__' and \
        params.get('output') or params.get('quiet')

    # evaluate() will automatically raise an exception if there are no tests

    for nr, (test, progress_pair) in enumerate(tester):
        if visual_notice:
            sys.stderr.write('%d ' % test['id'])
            sys.stderr.flush()
        # intotdeauna trebuie sa fie ori una ori alta
        if 'real_score' in test:
            real_score += test['real_score']
        else:
            score += test['score']
        nr_teste = nr + 1
        if display_table:
            #line = [line[i].ljust(lens[i]) for i in xrange(len(line))]
            '''
            order = ['Testul %(id)s', '%(verdict)s', '%(real_score)s', '%(time)s', '%(memory)s', '%(jrun_message)s']
            line = filter(None, map(lambda x: (x % test), order))
            '''
            order = ['id', 'real_score', 'time', 'memory', 'verdict'] # 'jrun_message'
            def adapt(str):
                if str == 'Time limit exceeded.':
                    return 'TLE'
                elif str == 'Wall time limit exceeded.':
                    return 'WTLE'
                elif str == 'Memory limit exceeded.':
                    return 'MLE'

            line = map(test.get, order)
            line[4] = test['verdict'] == 'FAIL' and adapt(test['jrun_message']) or line[4]
            line[0] = 'Testul %s' % line[0]

            # removal of empty fields, and making everything string
            line = [str(i) for i in line if i is not None]

            # give it table structure
            line = [j.ljust(lens[i]) for i, j in enumerate(line)]
            # Colorization
            if evalcore.output.isatty():
                line[-1] = col(test['highlight']) + line[-1] + col('normal')
                if 'real_score' in test:
                    line[1] = (test['highlight']=='green' and col('maialb') or '') + str(line[1]) + col('normal')
            log_print(''.join(line), 2)
            if 'debug_data' in test:
                path1 = self.options['nume'] + '-debug'
                path2 = '%d.diff' % test['id']
                try:
                    if not os.path.isdir(path1):
                        os.mkdir(path1)
                    open(join(path1, path2), 'wt').write(''.join(test['debug_data']))
                except (OSError, IOError), e:
                    print >>sys.stderr, ("Couldn't write debug data: " + str(e))

    if visual_notice:
        print >>sys.stderr

    puncte = real_score or int(round(score/nr_teste))
    log_print('Punctaj total: %d' % puncte, 1)

    if not display_table or params.get('output'):
        print puncte


def main(argv=None):
    'Parse parameters (argv or sys.argv) and evaluate one single executable'
    if argv is None:
        argv = sys.argv
    params, options = parse_parameters(argv)
    executabil = params['executabil']
    # Daca este specificat doar executabilul
    if not options.get('path') and not options.get('nume'):
        options['nume'] = os.path.basename(executabil)
    # Shortcut. Don't initialize object if it's just a one-time failed run
    # Problema.evaluate() makes this check this too
    if not isfile(executabil):
        die('The executable to be evaluated does not exist')
    evalcore.output = setup_output(quiet = params['quiet'], output_path = params['output'])

    problema = evalcore.Problema(params = params, configure = True, **options)
    #print_header(problema)
    #problema.open_config()
    print_results(problema, executabil, params)


if __name__ == "__main__":
    try:
        sys.exit(main())
    except evalcore.EvalError, e:
        print >>sys.stderr, e
    except IOError, e:
        print >>sys.stderr, e
    sys.exit(1)
