#!/usr/bin/env python

##############################################################
#                     (C) SANDULEAC DAN                      #
#                  EVALUATOR   version 0.64                  #
#                 RELEASED UNDER GPL LICENCE                 #
#============================================================#
#             this header is under construction              #
##############################################################

# For python 2.2 backwards compatibility, for yield
#from __future__ import generators

import sys, os, stat
import re
from operator import itemgetter
from os.path import join, isfile, isdir
import tempfile
# Shell utilities like copy, move etc
import shutil
from subprocess import Popen, PIPE

# CONSTANTS

_DIR_TESTE = '/home/dan/teste'

eval_options = {
    'chroot_command': '--chroot', # --copy-libs (don't use unless root)
    'args': [],
    'jrun': 'jrun',
}

defaults = {
    'time': 1000,
    'memory': 16384,
    'inf': '?-%s.in',
    'outf':'',
    'verif': ''
}

output = None

class EvalError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return (isinstance(self.value, str) \
            and self.value or str(self.value))

def die(message):
    'Raises an EvalError exception'
    raise EvalError(message)

#verbosity == 0  --- print only when writing to tty
#verbosity == 2  --- print even when writing to file

def log_print(message, verbosity = 1):
    ''' Puts a message to the file referenced by global variable `output`,
        if that exists
    '''
    # TODO: aici am de modificat conform eval_options['verbose']
    global output
    try:
        try:
            if verbosity == 0 and not output.isatty():
                return
        except AttributeError:
            return

        if output:
            print >>output, message
    except NameError: pass


# *** remember
# do something about afisarea stderr, ce facem cu ea daca suntem modul?

class Problema:

    def __init__(self, nume = None, path = None, params = {}, configure = False, **args):
        ''' Problema constructor
            params {} -> certain flags that help with configure(), such as repo
            also `executabil`, to help with standalone console format printing
            args -> problem specific properties, such as time, memory, nume, path, ...
        '''
        # Maybe split this in two pieces, excluding the automatic configuration ?

        if not nume and not path:
            raise EvalError('You must specify the name and/or the path, but not neither')
        self.punctaje = {}

        self.opts = defaults.copy() # shallow copy, but only simple objects here
        # 0.64 - changed here to explicit parameters
        # TODO: consider changing to self.nume, ...
        self.opts['nume'] = nume
        self.opts['path'] = path
        self.update(args)

        # FIXME: Poate ar trebui modificat si update(self) sa poata prelua quiet si output din
        # **args ... pana una alta ne multumim cu asta

        # facem recunoastere automata dupa name si path
        if configure or (not nume or not path):
            self.configure(params)

        # Continuing initialization procedure
        # Print_header won't do anything unless global open file 'output' is set
        self.print_header()
        self.open_config()
        self.sanity_check()

        if not self.opts['outf']:
            self.opts['outf'] = re.sub('.in$', '.ok', self.opts['inf'])


    def update(self, args):
        # perform a check, only allowed parameters? not sure if necessary
        for key, value in args.iteritems():
            if key not in defaults:
                die('Unknown parameter %s=%s' % (key, value))
            if not isinstance(value, type(defaults[key])):
                die('Wrong value type for %s = %s: correct is %s' % (key, value, str(type(defaults[key]))[7:-2]) )
            self.opts[key] = value
        return True
        #self.opts.update(args)


    def configure(self, params):
        ''' What this does is to try to deduce missing parameters from the already given ones
            Very nifty piece of code
            It is usually called automatically in the constructor
            ** Ported in version 0.63 from parse_parameters to help give the program modularity
        '''

        opts = self.opts


        # FIXME TODO: replace nume_tmp with opts['nume']
        # Reason:
        # Mai e nevoie de nume_tmp ? Path nu mai are voie sa fie "partial match" al numelui problemei
        # as of 0.62

        nume_tmp = opts['nume']
        # Auto-adapt name from executable if running in standalone mode.
        if not nume_tmp and 'executabil' in params:
            nume_tmp = re.sub('.*/', '', params["executabil"])

        # S-A DAT NUMELE ADEVARAT AL PROBLEMEI (SI EVENTUAL CALEA) prin -p
        if opts.get('path'):
            # Daca 'path' e o cale de director.. adaugam ``nume'' la el
            if opts["path"].endswith('/'):
                opts["path"] += nume_tmp
            else:
                # luam numele problemei din `path`, as opposed to din `executabil`
                nume_tmp = re.sub(r'^.*/', '', opts["path"])
        else: # compute path for problem's test directory from executable
            opts["path"] = nume_tmp

        # Now mess with dir_teste .. aka options.path
        # If can't find "%PATH" / "%PATH-teste" in current directory, search the repo
        pdirs = filter(isdir, [opts["path"], opts["path"] + '-teste'])
        if params.get("repo") or not pdirs:
            opts["path"] = find_dir_teste(_DIR_TESTE, opts["path"], 5)
        else:
            # we're explicitly marking that we're not using repo, because we found it here
            # CHANGED: why mark it? it's a useless flag anyway
            #params["repo"] = False
            opts["path"] = os.path.abspath(pdirs[0]) # Use first possible directory
        if not opts["path"]:
            die('Nu pot gasi directorul de teste')
        # Path poate fi incomplet, deci de-abia acuma luam numele din path, daca e cazul
        # FIXED: nume <- path doar daca am facut o cautare in repo
        if not opts["nume"]:
            opts["nume"] = params["repo"] and re.sub('.*/', r'', opts["path"]) or nume_tmp.lower()


    def open_config(self):
        opts = self.opts
        config = '%(path)s/%(nume)s.cfg' % opts
        PARAMETERS = ['inf', 'outf', 'time', 'memory', 'verif']
        if not isfile(config):
            try:
                config = find(opts['path'], '.*\.cfg$', type='f').next()[0]
            except StopIteration: pass
        if isfile(config):
            # Could do without it...
            log_print("Using config file `%s`" % os.path.basename(config), 0)
            f = open(config, 'r')
            for line in f:
                m = re.search(r'(\w+)\s*=\s*(\S+)', line)
                #print line
                if not m: continue
                key, value = m.groups()
                key = key.lower()
                if key in PARAMETERS and opts[key] is defaults.get(key):
                    # facem si typecasting ca sa nu-l ia string !!!
                    try:
                        opts[key] = type(opts[key])(value)
                        log_print('[config-file] %s=%s' % (key, value), 0)
                        # Verificatorul trebuie sa fie in directorul de configurare
                        if key == 'verif':
                            opts[key] = join(opts['path'], value)
                    except TypeError:
                        log_print("[config-file] Valoare incorecta pentru `%s', asteptam %s" % \
                                (key, type(opts[key])), 1)
                else:
                    if key not in PARAMETERS:
                        log_print('[config-file] Parametru necunoscut %s' % key, 1)
                    else:
                        log_print('[config-file] Parametru ignorat %s = %s' % (key,value), 1)
            f.close()

        if not opts['outf']:
            opts['outf'] = re.sub('.in$', '.ok', opts['inf'])
        # ==================================
        # fisierul tests.txt pentru punctaje
        # ==================================
        punc_file = join(opts['path'], 'tests.txt')
        if isfile(punc_file):
            try:
                for i in open(punc_file, 'r'):
                    id, score = i.split(' ')
                    self.punctaje[int(id)] = int(score)
            except (ValueError, KeyError), err:
                die('Eroare in fisierul tests.txt: %s' % str(err))


    def sanity_check(self):
        opts = self.opts
        if opts['verif']:
            if isfile(opts['verif']):
                opts['verif'] = os.path.abspath(opts['verif'])
                # HACK to accept pyc scripts here
                if opts['verif'].endswith('.pyc'):
                    opts['verif'] = ['python', opts['verif']]
                else:
                    opts['verif'] = [opts['verif']]
            else:
                print("Programul verificator `%s' nu exista, falling back to lazy diff" % \
                        opts['verif'])
                opts['verif'] = None


    def compara(self, input, output, ok):
        """ compara(input, output, ok) -> (mesaj, punctaj/punctaj_max %) """
        opts = self.opts

        if not opts['verif']:
            # just do the diff, lazy diff, ignore all trailing whitespace
            try:
                ln1 = [i.rstrip(' \n') for i in open(output, 'r')]
            except IOError:
                return 'Lipsa out', False
            ln2 = [i.rstrip(' \n\r') for i in open(ok, 'r')]
            if ln1 == ln2:
                return 'OK', 100
            else:
                return 'Gresit', 0
        else:
            #f, tmp = popen4('%s "%s" "%s" "%s"' % (opts['verif'], input, os.path.abspath(output), ok))
            command = opts['verif'] + [input, os.path.abspath(output), ok]
            f = Popen(command, stdout=PIPE).stdout
            ln = f.readlines()
            try:
                assert (len(ln) == 2)
                punctaj = int(ln[1])
            except:
                print >>sys.stdout, 'Dump - outputul verificatorului:'
                pprint(ln, stream=sys.stderr)
                die('Verificatorul nu urmeaza un format corect')
            assert(0 <= punctaj <= 100)
            exitcode = os.WEXITSTATUS(f.close() or 0)
            if exitcode is 0:
                return ln[0].rstrip('\n'), punctaj
            else:
                return 'Verificator: $?!=0', 0

    def prerun_check(self, chroot = False):
        # shall we chroot ?
        jrun_uid = os.stat(os.popen('which "%(jrun)s"' % eval_options).readline()[:-1])[stat.ST_UID]
        if eval_options.get('chroot'):
            if jrun_uid == 0 or os.getuid() == 0:
                eval_options['args'].append(eval_options['chroot_command'])
            else:
                die('Nu pot chroot, nu esti root si jrun nu are setuid root')


    def evaluate(self, executabil, debug = False, chroot = False):
        # Preliminary check
        executabil = os.path.abspath(executabil)
        if not isfile(executabil):
            die('The executable to be evaluated does not exist')
        self.prerun_check(chroot = chroot)
        return self.run_tests(executabil, debug)


    def run_tests(self, executabil, debug = False):
        ''' Yields (detalii, (test_crt, nr_teste))
        '''

        opts = self.opts
        workdir = tempfile.mkdtemp()
        opts['cwd'] = os.getcwd()
        
        try:
            try:
                shutil.copy(executabil, join(workdir, opts['nume']))
            except IOError, err:
                die(str(err))
            user_output = '%s.out' % opts['nume']

            # Ce parsari automate facem la inf si outf, facem numai
            # in contextul dictionarului test_files !
            test_files = { 'inf': opts['inf'], 'outf': opts['outf'] }
            if not test_files['outf']:
                test_files['outf'] = re.sub('.in$', '.ok', test_files['outf'])
            rules = (
                ('%s', opts['nume']),
            )
            for k in 'inf', 'outf':
                for old, new in rules:
                    test_files[k] = test_files[k].replace(old, new)
                test_files[k] = re.escape(test_files[k])
                test_files[k] = re.sub(r'\\\?+', '(\d+)', test_files[k])
            
            files_in = [(int(i[1]), i[0]) for i in iter( \
                find(opts['path'], test_files['inf'], type='f') )]
            files_out = dict((int(i[1]), i[0]) for i in iter( \
                find(opts['path'], test_files['outf'], type='f') ))
            #tests_d = dict( ( i[0], (i[1], files_out.get(i[0])) ) \
            #        for i in files_in )
            #tests = sorted(tests_d.items(), key = itg(0))
            tests = sorted((( tcrt[0], (tcrt[1], files_out.get(tcrt[0])) ) \
                    for tcrt in files_in),  key = itemgetter(0))

            if not tests:
                die('Nu am gasit niciun test in directorul de teste. Format teste asteptat: `%s` si `%s`' % (opts['inf'], opts['outf']))

            nr_teste = len(tests)

            for test_crt, (test, (inf, outf)) in enumerate(tests):
                detalii = { 'id': test, 'verdict': 'Eroare necunoscuta', 'highlight': 'else',
                    }
                os.chdir(workdir)

                if not outf and not opts['verif']:
                    detalii['verdict'] = 'Nu are .ok'
                else:
                    unixize(inf, '%s.in' % opts['nume'])
                    if isfile(user_output):
                        os.remove(user_output)
                    # TODO: supersede popen, use subprocess!
                    rez = os.popen(
                        eval_options['jrun'] + ' -p "%(nume)s" -m %(memory)d -t %(time)d ' % opts \
                        + ' '.join(eval_options['args'])).readline().strip().split(': ')
                    # Parsing jrun's output nicely
                    time, memory = re.match('time\s+(\w+)\s+memory\s+(\w+)', rez[1], re.I).groups()
                    detalii['time'] = time
                    detalii['memory'] = memory
                    #mesaj.extend([time, memory] + rez[2:])
                    detalii['verdict'] = rez[0]
                    detalii['jrun_message'] = rez[2]

                # daca a rulat cu succes, testam daca nu cumva da `Gresit'
                score = 0
                if detalii['verdict'] == 'OK':
                    msg, score = self.compara(input = inf, output = user_output, ok = outf)
                    detalii['verdict'] = msg
                    if score < 100 and isfile(user_output) and debug:
                        # Maybe cleanup a bit here, use tmp = proc.communicate('\n'.join(unixize_iter(outf)))[0]
                        command = ['diff', '-pruN', user_output, '-']
                        proc = Popen(command, stdout=PIPE, stdin=PIPE)
                        fo, fi = proc.stdin, proc.stdout
                        fo.writelines(unixize_iter(outf))
                        fo.close()
                        detalii['debug_data'] = fi.readlines() #''.join(fi)
                        fi.close()
                detalii['highlight'] = score == 0 and 'red' or \
                        score == 100 and 'green' or \
                        'green_i'
                if len(self.punctaje) > 0:
                    punctaj = int(round(self.punctaje.get(test, 10) * score / 100.0))
                    detalii['real_score'] = punctaj
                else:
                    detalii['score'] = score

                os.chdir(opts['cwd'])
                yield (detalii, (test_crt, nr_teste))

            try:
                shutil.rmtree(workdir)
            except Exception, e:
                log_print('Notice: nu am putut sterge workdir ' + workdir + ': ' + e, 0)

        finally:
            os.chdir(opts['cwd'])
    
    def print_header(self):
        #log_print("Evaluez `%s' folosind directorul `%s'" % (executabil, self.opts['path']), 0)
        log_print("Evaluez problema `%s'" % self.opts['path'], 0)
        log_print("Numele problemei: %s" % self.opts['nume'], 0)


def find(root, name, depth = 1, icase = True, type = 'df'):
    """Does a re.match() to test `name' against file_name"""
    dirs = [[]] * 2
    if not isdir(root): return
    dirs[0] = [root]
    for nv in xrange(depth):
        dirs[1] = []
        for dir in dirs[0]:
            if nv < depth-1:
                dirs[1].extend([x for x in
                    [join(dir,file) for file in os.listdir(dir)]
                    if isdir(x)]
            )
            for file in os.listdir(dir):
                m = re.match(name, file, icase and re.I)
                if not m: continue
                fullpath = join(dir, file)
                tip = isfile(fullpath) and 'f' or 'd'
                if tip in type:
                    yield (fullpath, m.groups() and m.groups()[-1])
        dirs[0] = dirs[1]

# Caut dir teste dupa root+path1, path2=ignorecase
def find_dir_teste(root, path, depth = 1):
    """Wrapper for find; Caut dir teste dupa root+path1, path2=ignorecase
    path = (.*/)?(.*) -> (path1, path2)"""
    path1, path2 = re.match("(.*/)?(.*)", path).groups()
    # Changed in 0.62
    # Accept -teste suffix too and DON'T allow partial problem name matches
    path2 += "(-teste)?$"
    try:
        results = find(join(root, path1 or ''), path2, depth=depth, type = 'd')
        # FIXME: what if there are more results!
        return results.next()[0]
    except StopIteration:
        return None

def unixize_iter(a):
    "Dos2unix done in python"
    f = open(a, 'rt')
    for line in f:
        yield line.replace('\r', '')

def unixize(a, b):
    "Dos2unix done in python"
    f = open(a, 'rt')
    b = open(b, 'wt')
    for line in f:
        b.write(line.replace('\r', ''))
